module WriteTypes {
  struct a {
    octet k[3];
    unsigned long long ll;
  };
#pragma keylist a k

  struct b {
    unsigned short k[3];
    unsigned long long ll;
  };
#pragma keylist b k

  struct c {
    unsigned long k[3];
    unsigned long long ll;
  };
#pragma keylist c k
  
  struct d {
    unsigned long long k[3];
    unsigned long long ll;
  };
#pragma keylist d k
};
