/*
 * Copyright(c) 2006 to 2018 ADLINK Technology Limited and others
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0, or the Eclipse Distribution License
 * v. 1.0 which is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 *
 * SPDX-License-Identifier: EPL-2.0 OR BSD-3-Clause
 */
module Space {
    struct Type1 {
	long	long_1; //@Key
	long	long_2;
	long	long_3;
    };
#pragma keylist Type1 long_1
    struct Type2 {
	long	long_1; //@Key
	long	long_2;
	long	long_3;
    };
#pragma keylist Type2 long_1
    struct Type3 {
	long	long_1;
	long	long_2;
	long	long_3;
    };
#pragma keylist Type3

    struct simpletypes {
        long                l;
        long long           ll;
        unsigned short      us;
        unsigned long       ul;
        unsigned long long  ull;
        float               f;
        double              d;
        char                c;
        boolean             b;
        octet               o;
        string              s; //@Key
    };
#pragma keylist simpletypes s
};
